require "Warshall.rb"
require "Grammatiken.rb"

# Author::    Jens Bachmann  (mailto:jens.bachmann@unibw.de)
# Author::		Daniel Borkowitz (mailto:daniel.borkowitz@unibw.de)
# Author::		Stefan Mayer (mailto:stefan.mayer@unibw.de)
# Author::		Roman Krhling (mailto:roman.kraehling@unibw.de)
# Copyright:: Copyright (c) 2006 The Heimschlfers, UniBw Mnchen

# Diese Klasse stellt ein Item eines Item-Automaten dar, z.B. [S->X.Ya]. (S.154), bzw fr LR(1) usw. noch zustzlich die follow-Menge
class Item
	# Konstruktor bekommt einen String z.B. [S -> .E + T ] (Item) 
	# und einen Rechtskontext als Menge von Strings
	def initialize(string,rechts)
		@item=string
		@rechtskontext = rechts
	end
	
	# Setter fr Rechtskontext
	def setRechts!(rechts)
		@rechtskontext = rechts
		return self
	end

	# Getter fr Rechtskontext
	def getRechtskontext
		return @rechtskontext
	end
	
	# liefert String z.B. [S -> .E + T ] (Item) 
	def getItem
		return @item
	end
	
	# liefert die linke Seite des Items als String
	def get_LinksSymbol
		return @item.split[0]
	end
	
	# Verknpft zwei Itemstrings
	def +(it)
	  return Item.new(@item + it.getItem, @rechtskontext)
	end
	
	# liefert eine wahre Kopie von Item
  def copy
	 return Item.new(@item, @rechtskontext.copy)
	end

	# setzt den Punkt um ein Symbol weiter nach rechts
	def lesen
		ind = getItem.index('.')
		substring = @item[ind+1..(@item.length-1)]
		ch = substring.split[0]
		return nil if substring[(ch.length+1)..substring.length-1]==nil
		return Item.new(@item[0..ind-1]+ch+" ."+substring[(ch.length+1)..substring.length-1],@rechtskontext)
	end 
 
	# prft ob mit gegebenem Symbol gelesen werden kann
  def lesen?(symb)
	 	ind = getItem.index('.')
		substring = @item[ind+1..(@item.length-1)]
		ch = substring.split[0]
	  return true if symb.to_s==ch
		return false
	end 

	# prft ob der Punkt am Ende angekommen ist
	def ist_ende
		return @item[@item.length-1,1]=="."
	end
	
	# liest das Item um ein Symbol rckwrts
	def rueck_lesen(symb)
	  str = getItem.reverse
		ind = str.index('.')
		substring = str[ind+2..(str.length-1)]
		ch = substring.split[0]
		return nil if ((substring[0,1]=='>') or (ch.reverse!=symb.to_s))
		return Item.new((" "+ch+". "+substring[(ch.length+1)..substring.length-1]).reverse,@rechtskontext) if ind==0
		return Item.new((str[0..ind-1]+" "+ch+". "+substring[(ch.length+1)..substring.length-1]).reverse,@rechtskontext)
	end 
 
	# Vergleich zweier Items
	def <=>(anOther)
		return @item<=>anOther.getItem
	end
	
	# Tiefenvergleich zweier Items
	def ==(anOther)
		return ((@item == anOther.getItem)and (@rechtskontext == anOther.getRechtskontext))  if anOther!=nil
		return false
	end
	
	# Vergleich zweier Items
	def ===(anOther)
		return (@item == anOther.getItem) if anOther!=nil
		return false
	end
	
	# gibt ein Ausgabestring von Item zurck
	def to_s
		return "[" + @item + "]"
	end
	
	# gibt ein Ausgabestring von Item mit Rechtskontext zurck
	def to_s_rechts
		return "[" + @item + "]" if @rechtskontext==nil
		return "[" + @item + ";" + @rechtskontext.to_s + "]"
	end
	
	# liefert das nchste Zeichen nach dem Punkt
	def naechstesZeichen
		ind = getItem.index('.')
		substring = @item[ind+1..(@item.length-1)]
		return substring.split[0]
	end
	
	# liefert den Rest des Items dem Punkt und einem Folgesymbol als Menge von Strings
	# [S -> .E + T] liefert {"+","T"}
	def naechsterRest
		ind = getItem.index('.')
		substring = @item[ind+1..(@item.length-1)]
		ch = substring.split[0]
		return nil if ch == nil
		str=@item[ind+ch.length+1..@item.length-1]
		erg = Menge.new([])
		str.split.each{|s| erg.add(s)}
		return erg
	end

	# prft ob das Folgezeichen des Punktes ein NonTerminal ist
	def isEpsUebergang(gram)
		return gram.getNonterminals.include?(naechstesZeichen)
	end
	
	# prft ob der Zustand der Ausgangszustand von einem NonTerminal a ist
	def ist_start_von(a)
		ind = getItem.index('.')		
		return ((@item[ind-2,1]) == ">" and (get_LinksSymbol == a))
	end
end

# Diese Klasse stellt das Tripel einer LR-Konfiguration dar (S.155)
# [Zustand,bergangssymbol,Zustand]
class Konfiguration
	
	# Konstruktor bekommt einen Zustand, ein bergangssymbol und Folgezustnde
	def initialize(zustand, symbol, folge)
		@zustand = zustand
		@symbol = symbol
		@folge = folge
	end
	
	# getter fr die Komponenten des Tripels(0-2)
	def [](i)
	 case i
	    when 0 then @zustand
			when 1 then @symbol
			when 2 then @folge
			else nil
		end
	end
	
	# Vergleich zweier Konfigurationen
	def ==(anOther)
	  if anOther != nil then
			return (@zustand==anOther[0] and @symbol==anOther[1] and @folge==anOther[2]) 
		else 
			return false
		end
	end
	
	# Vergleich zweier Konfigurationen
	def <=>(anOther)
		return @zustand<=>anOther[0]
	end
	
	# liefert einen Ausgabestring fr eine Konfiguration
	def to_s
		return ("[" + @zustand.to_s + " --" + @symbol + "--> " + @folge.to_s + "]" + "\n")
	end
	
	# liefert einen Ausgabestring fr eine Konfiguration mit Rechtskontexten
	def to_s_rechts
		return ("[" + @zustand.to_s_rechts + " --" + @symbol + "--> " + @folge.to_s_rechts + "]" + "\n")
	end

	# liefert den Folgezustand eines Zustands nach "lesen" eines Symbols
	def folgeZustand?(zustand,symbol)
		if ((zustand === @zustand) and (@symbol == symbol)) then
			tmp = @folge.copy
			return tmp
		else 
			return nil
		end
	end	
	
	# berechnet den Startzustand fr den LR_1
	def berechne_LR_1_Start(zustand,symbol,grammatik)
		parser=ParserRelation.new(grammatik)
		if ((zustand === @zustand) and (@symbol == symbol)) then
			if symbol != "eps" then
				folge=@folge.getNEA_item.copy
				folge.setRechts!(zustand.getNEA_item.getRechtskontext)
				return Zustand.new(folge)
			else
				folge = @folge.getNEA_item.copy
				folge.setRechts!(parser.FIRST(zustand.getNEA_item.naechsterRest,zustand.getNEA_item.getRechtskontext)+zustand.getNEA_item.getRechtskontext)
				return Zustand.new(folge)
			end
		end
	  return nil
	end	
	
end

# Beschreibt einen Zustand des Item-Automaten, also eine Menge von Items
class Zustand
	# Konstruktor bekommt ein Item
	def initialize (item)
		@items = Menge.new([])
		@items.add(item) if item != nil
	end
	
	# liefert eine Tiefenkopie
	def copy
		temp=Zustand.new(nil)   
		@items.each{|i| temp.add_i(i.copy) if i != nil }
		return temp
	end
	
	# fgt dem Zustand ein Item hinzu
	def add_i(item)
		@items.add(item) if item != nil
		return self
	end
	
	# fgt dem Zustand einen anderen Zustand bzw. dessen Items hinzu, vgl Vereinigung
	def add_Z(zustand)
		@items+=zustand.items if zustand != nil
		return self
	end
	
	# liefert die Items des Zustands als String zurck
	def to_s
		@items.to_s
	end
	
	# liefert die rechte Seite des Zustands als String zurck
	def to_s_rechts
		str = "{ "
		each{|i| str=str+i.to_s_rechts+", "}
		str = str[0, str.length()-2]
		str+= " }"
		return str
	end
	
	# getter fr @items
	def items
		@items
	end
	
	# getter fr das erste Item
	def getNEA_item
		@items[0]
	end
	
	# Vergleich zweier Zustnde, liefert true oder false zurck
	def ==(another)
		if another!=nil then 
			return (@items == another.items) 
		else 
			return false 
		end
	end
	
	# Tiefenvergleich
	def ===(another)
		if another != nil then 
			for i in 0..(@items.length-1) do
				return false if !(@items[i]===another.items[i])
			end
			return true
		else 
			return false 
		end
	end
	
	# Vergleich zweier Zustnde, liefert -1, 0 oder 1 zurck
	def <=>(anOther)
		return @items <=> anOther.items
	end
	
	# prft die Mengen der Items beider Zustnde auf Gleichheit,
	# bei Gleichheit werden die rechten Seiten zusammengefasst
	def einSortieren(zustand)
		tmp = zustand.getNEA_item
		@items.each{|i|
		    if i === tmp then
						i.setRechts!(i.getRechtskontext+tmp.getRechtskontext) 
						return self
				end
		}
    @items += zustand.items
		return self
	end
	
	# sortiert das @items-array 
	def sort!
		@items.sort!
		return self
	end
	
	# liefert den Folgezustand des Zustands mit einem Symbol und einem bergang zurck
	def folgeZustand(symbol, uebergang)
		temp = Zustand.new(nil)
	  each{|i| 
			  uebergang.each{ |u| 
						folge = u.folgeZustand?(Zustand.new(i), symbol)
						temp.add_Z(folge) if folge!=nil
				}
		}
		while true do
			print "0"
			vergl = temp.copy
			temp.each{|t|
					uebergang.each{ |u| 
							folge = u.folgeZustand?(Zustand.new(t), "eps")
							temp.add_Z(folge) if folge!=nil
				 }
			}
			break if vergl==temp
		end
		return temp
	end
	
	# berechnet den Startzustand des LR(1) Automaten
	def berechne_LR_1_Start(symbol, uebergang,grammatik)
		temp=Zustand.new(nil)
	  each{|i| 
			  uebergang.each{ |u| 
						folge = u.berechne_LR_1_Start(Zustand.new(i), symbol,grammatik)
						temp.einSortieren(folge) if folge != nil
				 }
		}
		while true do
			print "1"
			vergl = temp.copy
			temp.each{|t|
					uebergang.each{ |u| 
							folge = u.berechne_LR_1_Start(Zustand.new(t), "eps",grammatik)
							temp.einSortieren(folge) if folge!=nil 
					}
			}
			break if vergl==temp
		end
		return temp
	end

	# liefert das gelesene Item eines Zustandes
  def get_schalt_It(item,symb)
	  self.each{|i|
				if item.rueck_lesen(symb)!=nil then 
					return i if i===item.rueck_lesen(symb) 
				end
		}
		return nil
	end

	# liefert den Folgezustand eines Zustandes eines LR(1)-Automaten
	def folgeZustand_LR(symbol,uebergang,grammatik)
		tmpit = nil
		parser = ParserRelation.new(grammatik)
		print "."
		folge=Zustand.new(nil)
	  uebergang.each{ |u|
			 folge=u.folgeZustand?(self,symbol) if u.folgeZustand?(self,symbol)!=nil
		}
		if folge !=nil
			folge=folge.copy
					folge.each{|item|
						temp_item_geschalten = self.get_schalt_It(item,symbol)
						if temp_item_geschalten != nil 
						then 
						item.setRechts!(temp_item_geschalten.getRechtskontext().copy())
						else item.setRechts!(Menge.new([]))
						end
					}
		tmpfolge=Zustand.new(nil)
		while tmpfolge!=folge do
			tmpfolge=folge.copy
			folge.each{|item|
				if item.isEpsUebergang(grammatik)	then 
					folge.each{|it|
						if it.ist_start_von(item.naechstesZeichen) then 
							it.setRechts!(parser.FIRST(item.naechsterRest, item.getRechtskontext)+it.getRechtskontext)
						end
					}
				end
			}
		end
   end
		return folge if folge!=Zustand.new(nil)
		return nil
end

	# Iterator ber @items
	def each &b
		@items.each &b
	end
end

# Diese Klasse stellt einen Nichtdeterministischen Item-Automaten zu einer Grammatik dar.
class NEA
	
	# der NEA besteht aus einer Menge von Zustnden, einem Startzustand und einer Menge von 
	# Konfigurationen als bergangsfunktion.
	def initialize(grammatik)
		@grammar = grammatik
		start = @grammar.getStartsymbol.getName
		@zustaende = Menge.new([])
		@alphabet_term = Menge.new([])
		@alphabet = Menge.new([])
		@grammar.getTerminals.each_value{|t| @alphabet.add(t.to_s)}
		@alphabet_term = @alphabet.copy
		@alphabet_term.add("#")
		@grammar.getNonterminals.each_value{|t| @alphabet.add(t.to_s)}
		@startZustand = Zustand.new(Item.new("S' -> ."+start+" ",@alphabet_term)) 
		berechneZustaende(@grammar.addRule("S' -> "+start+"\n\n"))
		@uebergaenge = uebergangsfunktion(@grammar.addRule("S' -> "+start+"\n\n"))
	end

	# Getter fr die Grammatik des Automaten
	def getGrammar
	 @grammar
	end

	# berechnet die Zustnde des Automaten als Menge
	def berechneZustaende(grammar)
		grammar.getRules.each_value do |x| 
				links = x.getSymbol
				tmpMenge = Menge.new(x.getAlts)
				tmpMenge.each do |i| 
					for b in 0..i.getSymbols.length do
						str=Item.new(links.to_s+" -> ",@alphabet_term)
						for c in 0..i.getSymbols.length-1 do
							str+=Item.new(".",nil) if ((c==b))
							str+=Item.new(i.getSymbols[c].to_s+" ",nil) 
						end
						str+=Item.new(".",nil) if (b==i.getSymbols.length)
						temp = Zustand.new(str)
						@zustaende.add(temp)
					end
				end
		end
	end
	
	# berechnet die bergnge des Automaten als Menge von LR-Konfigurationen
	def uebergangsfunktion(grammar)
		erg = Menge.new([])
		@zustaende.each do |z| 
				ch = z.getNEA_item.naechstesZeichen
				if (z.getNEA_item.isEpsUebergang(grammar)) then
					@zustaende.each do |folg| 
						erg.add(Konfiguration.new(z,"eps",folg)) if  folg.getNEA_item.ist_start_von(ch.to_s) 
					end
				end
				erg.add(Konfiguration.new(z,ch,Zustand.new(z.getNEA_item.lesen))) if ch != nil
		end
		return erg 
	end
			
	# liefert die bergaenge des NEA		
	def getUebergaenge
		@uebergaenge
	end
	
	# liefert die Zustaende des NEA
	def getZustaende
	 @zustaende
	end
	
	# liefert den Startzustand des NEA
	def getStartZustand
		@startZustand
	end
	
	# Setter fuer die Uebergaenge des NEA
	def setUebergaenge(ueb)
		@uebergaenge = ueb
		return self
	end
	
	# Setter fuer die Zustaende des NEA
	def setZustaende(zust)
		@zustaende = zust
		return self
	end
	
	# Setter fuer den Startzustand des NEA
	def setStartZustand(st)
		@startZustand = st
		return self
	end
	
	# Liefert eine wahre Kopie des NEA
	def copy
		temp= NEA.new(Grammar.new(""))
		temp.setUebergaenge(@uebergaenge.copy)
		temp.setZustaende( @zustaende.copy)
		temp.setStartZustand(@startZustand.copy)
		return temp
	end
	
	# Speichert das Objekt NEA auf der Harddisk mit Dateiname als Parameter mittels Marshall	
	def speichern(dateiname)
		File.open(dateiname, "w") do |f|
			Marshal.dump(self,f)
		end
	end
	
	# Ldt das Objekt NEA von der Harddisk mit Dateiname als Parameter mittels Marshall	
	def laden(dateiname)
		File.open(dateiname, "r") do |f|
			Marshal.load(f)
		end
	end
	
	# Formatierte Ausgabe des Automaten als String
	def to_s
		str = "\n\nDer " + self.class.to_s + "-Automat " + "hat " + @zustaende.length.to_s + " Zustaende\n"
		str+= "\nStartzustand(Nr.0) enthaelt folgende Items: \n" +@startZustand.to_s_rechts + "\n\n"
				j = 1
				@zustaende.each{|i| 
				if i!=@startZustand then
				str+=("Zustand Nr."+j.to_s+" enthaelt folgende Items:\n" + i.to_s_rechts+"\n\n") 
				j+=1
				end}
		#str+= "\n\n Uebergaenge:\n "
		#	  @uebergaenge.each{|i| str+= ("\t\t"+i.to_s_rechts+"\n")}
		str+="\n\n"
		return str
	end
		
	# each-Iterator fr @zustaende
	def zustaende_each &b
	 @zustaende.each &b
	end
	
	# each-Iterator fr @uebergaenge
	def uebergaenge_each &b
	 @uebergaenge.each &b
	end
end

# Klasse des LR(0) Automaten abgeleitet von NEA
class LR_0<NEA
	
	# super-Konstruktor Aufruf
	def initialize(grammatik)
		super(grammatik)
		@grammar = grammatik
		berechneStartZustand
		myhill
		uebergaenge
	end
	
	# Getter fr die Grammatik des Automaten
	def getGrammar
	 @grammar
	end
	
	# berechnet den Startzustand des LR(0)
	def berechneStartZustand
		start = @startZustand.copy
		uebergaenge = @uebergaenge.copy
		@startZustand = Zustand.new(nil)
		@startZustand.add_Z(start)
		@startZustand.add_Z(start.folgeZustand("eps",uebergaenge))
	end

	# Myhillsche Teilmengenkonstruktion des zugehrigen DEA S.157
	def myhill
		uebergaenge = @uebergaenge.copy
		zustaende = @zustaende.copy
		@zustaende = Menge.new([])
		@zustaende.add(@startZustand)
		tmpZustand = Menge.new([])
		while tmpZustand!=@zustaende do
			tmpZustand = @zustaende.copy
			tmpZustand.each{|i| 
						@alphabet.each{ |z| 
									folge = i.folgeZustand(z,uebergaenge)
									@zustaende.add(folge) if folge != Zustand.new(nil)
						}
			}
		end
	end

	# berechnet die Uebergangsfunktion des LR(0)
	def uebergaenge
		tempueb = @uebergaenge.copy
		@uebergaenge = Menge.new([])
		@zustaende.each{ |z|
				@alphabet.each{ |a|
						folge = z.folgeZustand(a,tempueb)
						if folge != Zustand.new(nil) then
							@uebergaenge.add(Konfiguration.new(z, a, folge)) if a != "eps"
						end
				 }
			}
	end
	
	# Ausgabe mit Rechtskontext
	def to_s_rechts
		str = "\nStartzustand: " + @startZustand.to_s_rechts + "\n\n"
		str+= "Zustaende: ("+@zustaende.length.to_s+") \n "
		@zustaende.each{|i| str+=("\t\t"+i.to_s_rechts+"\n")}
		str+= "\n\n Uebergaenge:\n "
		@uebergaenge.each{|i| str+= ("\t\t"+i.to_s_rechts+"\n")}
		str+="\n\n"
		return str
	end

	# Formatierte Ausgabe des Automaten als String
	def to_s
		str = "\n\nDer " + self.class.to_s + "-Automat " + "hat " + @zustaende.length.to_s + " Zustaende\n"
		str+= "\nStartzustand(Nr.0) enthaelt folgende Items: \n" +@startZustand.to_s_rechts + "\n\n"
				j = 1
				@zustaende.each{|i| 
				if i!=@startZustand then
				str+=("Zustand Nr."+j.to_s+" enthaelt folgende Items:\n" + i.to_s_rechts+"\n\n") 
				j+=1
				end}
		str+= "\n\n Uebergaenge:\n "
			  @uebergaenge.each{|i| str+= ("\t\t"+i.to_s_rechts+"\n")}
		str+="\n\n"
		return str
	end

end

# Klasse fuer den SLR(1) Automaten abgeleitet von LR(0)
class SLR_1<LR_0

	# Super Konstruktor
	def initialize(grammatik)
		start = grammatik.getStartsymbol.getName
		super
		@grammar = grammatik
		mache_SLR_1
	end
	
	# Getter fr die Grammatik des Automaten
	def getGrammar
	 @grammar
	end

	# veraendert den LR(0) und berechnet die Rechtskontexte
	def mache_SLR_1
		parser = ParserRelation.new(@grammar)
		@startZustand.each{|it|  
				it.setRechts!(parser.FOLLOW(it.get_LinksSymbol)) 
				it.setRechts!(Menge.new(["#"])) if it.get_LinksSymbol=="S'"														 								 
		}
	  zustaende_each{|z|
				z.each{|it|  
						it.setRechts!(parser.FOLLOW(it.get_LinksSymbol)) 
						it.setRechts!(Menge.new(["#"])) if it.get_LinksSymbol=="S'"														 								 
							}
		}
	end

end

# Klasse fuer den Automaten LR(1) abgeleitet von LR(0)
class LR_1<LR_0

	# Superkonstruktor
  def initialize(grammatik)
		super(grammatik)
		@grammar = grammatik
		@kernZustaende= @zustaende
		@zustaende=Menge.new([])
		@zustaende.add(@startZustand)
		myhill1
	end

	# Getter fr die Grammatik des Automaten
	def getGrammar
	 @grammar
	end

	# berechnet den Startzustand des LR(1)
	def berechneStartZustand()
		start = @startZustand.copy
		uebergaenge = @uebergaenge.copy
		@startZustand.add_Z(start)
		@startZustand.getNEA_item.setRechts!(Menge.new(["#"]))
		@startZustand.add_Z(@startZustand.berechne_LR_1_Start("eps",uebergaenge,@grammar))
	end

	# Myhillsche Teilmengenkonstruktion des zugehrigen DEA
	def myhill1()
		uebergaenge=@uebergaenge.copy
		@uebergaenge=Menge.new([])
	  tmpZustand =Menge.new([])
		while (tmpZustand.length!=@zustaende.length) do
			   tmpZustand=@zustaende.copy
				 tmpZustand.each{|i| 
							@alphabet.each{ |a| 
									folge = i.folgeZustand_LR(a,uebergaenge,@grammar)
									if folge !=nil
									   @zustaende.add(folge)
									   @uebergaenge.add(Konfiguration.new(i,a,folge))
									end
							}
					}
		end
	end		
end

class ParserRelation

	def initialize(grammar)
		@grammar = grammar	
		@start = @grammar.getStartsymbol.getName
	end

	def grammar
		@grammar
	end
	
	def addRule(rule)
		@grammar = @grammar.addRule(rule)
	end
	
	# liefert die Relation zur BEGIN Funktion zu einer Grammatik
	def BEGIN
		tmpRel = Relation.new([[]])
		@grammar.getTerminals.each_value{|i| tmpRel.addPaar!(i.to_s,i.to_s) }
		@grammar.getRules.each_value do |x| 
			if !(x.isEpsilon) then
				links = x.getSymbol 
				tmpMenge = Menge.new(x.getAlts)
				tmpMenge.each do |i| 
					tmpRel.addPaar!(links.to_s, i.getSymbols[0].to_s)
					j=0
					while ((j<i.getSymbols.length-1)and(@grammar.getEpsilonSymb.include?(i.getSymbols[j]))) do
						j+=1
						tmpRel.addPaar!(links.to_s, i.getSymbols[j].to_s ) 
					end
				end
			end
		end
	return tmpRel
	end
	
	
	# liefert die Relation zur END Funktion zu einer Grammatik
	def END
		tmpRel = Relation.new([[]])
		@grammar.getRules.each_value do |x| 
			if !(x.isEpsilon) then
				links = x.getSymbol 
				tmpMenge = Menge.new(x.getAlts)
				tmpMenge.each do |i| 
					j=i.getSymbols.length-1
					tmpRel.addPaar!(links.to_s, i.getSymbols[j].to_s)
					while ((j>0)and(@grammar.getEpsilonSymb.include?(i.getSymbols[j]))) do
						j-=1
						tmpRel.addPaar!(links.to_s, i.getSymbols[j].to_s ) 
					end
				end
			end
		end
	return tmpRel
	end
	
	def NEXT
		tmpRel = Relation.new([[]])
		@grammar.getRules.each_value do |x| 
			if !(x.isEpsilon) then
				links = x.getSymbol
				tmpMenge = Menge.new(x.getAlts)
				tmpMenge.each do |i| 
					for b in 0..(i.getSymbols.length-2)
						j=1
						while ((b+j<i.getSymbols.length)and(@grammar.getEpsilonSymb.include?(i.getSymbols[b+j]))) do
							tmpRel.addPaar!(i.getSymbols[b].to_s, i.getSymbols[b+j].to_s )
							j+=1
						end
						if (b+j<i.getSymbols.length) then tmpRel.addPaar!(i.getSymbols[b].to_s, i.getSymbols[b+1].to_s ) end
					end
				end
			end
		end
		return tmpRel
	end

	# liefert die FIRST Menge fr eine Grammatik 
	# zu einer gegebenen Zeichenkette von Nonterminalen
	def FIRST(a,b)
		addRule("S' -> "+@start+" #\n\n")
		erg = Menge.new([])
		a.each{|symb|
			erg = erg + self.BEGIN.reflexiv.transitiv.entferneNonterminale.nachfolger(symb)
			if !(@grammar.getEpsilonSymb.include?(Symb.new(symb))) then 
				return erg.copy
			end
			}
		# zur Berechnung LR_1 bentigt Rechtskontext wird vereinigt 
		# wenn alles Epsilonsymbole vorher vorhanden
		erg += b if b != nil
		return erg
	end
	
	# druckt wie Jaccie alle FirstSets der Grammatik aus
	def FirstSets
		firstString = String.new
		@grammar.getRules.each_value do |x| 
			links = x.getSymbol
			tmpMenge = Menge.new(x.getAlts)
			tmpMenge.each do |i| 
				firstString = firstString + links.to_s+" -> "+i.to_s+" "+FIRST(i.to_s,nil).to_s + "\n"
			end
		end
		return firstString
	end

	# liefert die FOLLOW Menge fr eine Grammatik 
	# zu einer gegebenen Zeichenkette von Nonterminalen
	def FOLLOW(a)
		addRule("S' -> "+@start+" #\n\n")
		erg = Menge.new([])
			erg = erg + (self.END.inv.reflexiv.transitiv.komp(self.NEXT).komp(self.BEGIN.reflexiv.transitiv)).entferneNonterminale.nachfolger(a)
		return erg
	end

	# druckt wie Jaccie alle FollowSets der Grammatik aus
	def FollowSets
		followString = String.new
		@grammar.getRules.each_value do |x| 
			links = x.getSymbol
			followString = followString + links.to_s+"  "+FOLLOW(links.to_s).to_s+"\n"
		end
		return followString
	end

end									 
